@@ -9,7 +9,20 @@ setGeneric(name="all.equal", useAsDefault=all.equal)
 
 setGeneric(name="as.vector", 
   function(x, ...)
-    standardGeneric("as.vector")
+    standardGeneric("as.vector"),
+  package="pbdBASE"
+)
+
+setGeneric(name="rbind", 
+  function(..., ICTXT=0, deparse.level=1)
+    standardGeneric("rbind"),
+  package="pbdBASE"
+)
+
+setGeneric(name="cbind", 
+  function(..., ICTXT=0, deparse.level=1)
+    standardGeneric("cbind"),
+  package="pbdBASE"
 )
 
 ### S4 methods for new things
@@ -19,12 +32,6 @@ setGeneric(name="as.ddmatrix",
   package="pbdBASE"
 )
 
-setGeneric(name="as.vec", 
-  function(x, ...) 
-    standardGeneric("as.vec"), 
-  package="pbdBASE"
-)
-
 setGeneric(name="submatrix", 
   function(x, ...) 
     standardGeneric("submatrix"), 
@@ -33,7 +40,8 @@ setGeneric(name="submatrix",
 
 setGeneric("submatrix<-", 
   function(x, value)
-    standardGeneric("submatrix<-")
+    standardGeneric("submatrix<-"),
+  package="pbdBASE"
 )
 
 setGeneric(name="ldim", 
